# Halo mass definitions
from .massdef import (
    mass2radius_lagrangian,
    MassDef,
    MassDef200m,
    MassDef200c,
    MassDef500c,
    MassDefVir,
    convert_concentration,
)

# Halo mass-concentration relations
from .concentration import (
    Concentration,
    ConcentrationDiemer15,
    ConcentrationBhattacharya13,
    ConcentrationPrada12,
    ConcentrationKlypin11,
    ConcentrationDuffy08,
    ConcentrationIshiyama21,
    ConcentrationConstant,
    concentration_from_name,
)

# Halo mass functions
from .hmfunc import (
    MassFunc,
    MassFuncPress74,
    MassFuncSheth99,
    MassFuncJenkins01,
    MassFuncTinker08,
    MassFuncTinker10,
    MassFuncWatson13,
    MassFuncAngulo12,
    MassFuncDespali16,
    MassFuncBocquet16,
    mass_function_from_name,
)

# Halo bias functions
from .hbias import (
    HaloBias,
    HaloBiasSheth99,
    HaloBiasSheth01,
    HaloBiasTinker10,
    HaloBiasBhattacharya11,
    halo_bias_from_name,
)

# Halo profiles
from .profiles import (
    HaloProfile,
    HaloProfileGaussian,
    HaloProfilePowerLaw,
    HaloProfileNFW,
    HaloProfileEinasto,
    HaloProfileHernquist,
    HaloProfilePressureGNFW,
    HaloProfileHOD,
)

# Halo profile 2-point cumulants
from .profiles_2pt import (
    Profile2pt,
    Profile2ptHOD,
)

# Halo model power spectrum
from .halo_model import (
    HMCalculator,
    halomod_mean_profile_1pt,
    halomod_bias_1pt,
    halomod_power_spectrum,
    halomod_Pk2D,
    halomod_trispectrum_1h,
    halomod_Tk3D_1h,
    halomod_Tk3D_SSC,
    halomod_Tk3D_SSC_linear_bias,
)

# CIB profiles
from .profiles_cib import (
    HaloProfileCIBShang12,
    Profile2ptCIB,
)


__all__ = (
    'mass2radius_lagrangian', 'MassDef', 'MassDef200m', 'MassDef200c',
    'MassDef500c', 'MassDefVir', 'convert_concentration',
    'Concentration', 'ConcentrationDiemer15', 'ConcentrationBhattacharya13',
    'ConcentrationPrada12', 'ConcentrationKlypin11', 'ConcentrationDuffy08',
    'ConcentrationIshiyama21', 'ConcentrationConstant',
    'concentration_from_name',
    'MassFunc', 'MassFuncPress74', 'MassFuncSheth99', 'MassFuncJenkins01',
    'MassFuncTinker08', 'MassFuncTinker10', 'MassFuncWatson13',
    'MassFuncAngulo12', 'MassFuncDespali16', 'MassFuncBocquet16',
    'MassFuncBocquet20', 'mass_function_from_name',
    'HaloBias', 'HaloBiasSheth99', 'HaloBiasSheth01', 'HaloBiasTinker10',
    'HaloBiasBhattacharya11', 'halo_bias_from_name',
    'HaloProfile', 'HaloProfileGaussian', 'HaloProfilePowerLaw',
    'HaloProfileNFW', 'HaloProfileEinasto', 'HaloProfileHernquist',
    'HaloProfilePressureGNFW', 'HaloProfileHOD',
    'Profile2pt', 'Profile2ptHOD',
    'HMCalculator', 'halomod_mean_profile_1pt', 'halomod_bias_1pt',
    'halomod_power_spectrum', 'halomod_Pk2D', 'halomod_trispectrum_1h',
    'halomod_Tk3D_1h', 'halomod_Tk3D_SSC', 'halomod_Tk3D_SSC_linear_bias',
    'HaloProfileCIBShang12', 'Profile2ptCIB',
)



import numpy as np
import pyccl as ccl
import pytest

NZ = 128
ZZ = np.linspace(0., 1., NZ)
BZ_C = 2.
BZ = BZ_C * np.ones(NZ)
COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
TRS = {'TG': ccl.nl_pt.PTNumberCountsTracer((ZZ, BZ),
                                            (ZZ, BZ),
                                            (ZZ, BZ)),
       'TI': ccl.nl_pt.PTIntrinsicAlignmentTracer((ZZ, BZ),
                                                  (ZZ, BZ),
                                                  (ZZ, BZ)),
       'TM': ccl.nl_pt.PTMatterTracer()}
PTC = ccl.nl_pt.PTCalculator(with_NC=True,
                             with_IA=True,
                             with_dd=True)

a_1 = a_2 = a_d = 1.0
gz = ccl.growth_factor(COSMO, 1./(1+ZZ))
ZZ_1 = 1.0
gz_1 = ccl.growth_factor(COSMO, 1./(1+ZZ_1))
a_1_v = a_1*np.ones_like(ZZ)
Om_m = COSMO['Omega_m']
rho_crit = ccl.physical_constants.RHO_CRITICAL
rho_m = ccl.physical_constants.RHO_CRITICAL * COSMO['Omega_m']
Om_m_fid = 0.3

c_1_t = -1*a_1*5e-14*rho_crit*COSMO['Omega_m']/gz
c_1_t_1 = -1*a_1*5e-14*rho_crit*COSMO['Omega_m']/gz_1
c_d_t = -1*a_d*5e-14*rho_crit*COSMO['Omega_m']/gz
c_2_t = a_2*5*5e-14*rho_crit*COSMO['Omega_m']**2/(Om_m_fid*gz**2)
c_2_t_des = a_2*5*5e-14*rho_crit*COSMO['Omega_m']/(gz**2)

ks = np.logspace(-3, 2, 512)


def test_pt_tracer_smoke():
    ccl.nl_pt.PTTracer()


def test_pt_tracer_m_smoke():
    ccl.nl_pt.PTMatterTracer()


@pytest.mark.parametrize('b2', [(ZZ, BZ), BZ_C, None])
def test_pt_tracer_nc_smoke(b2):
    pt_tr = ccl.nl_pt.PTNumberCountsTracer((ZZ, BZ),
                                           b2=b2,
                                           bs=(ZZ, BZ))

    # Test b1 and bs do the right thing
    for b in [pt_tr.b1, pt_tr.bs]:
        assert b(0.2) == BZ_C

    # Test b2 does the right thing
    if b2 is not None:
        assert pt_tr.b2(0.2) == BZ_C
        zz = np.array([0.2])
        assert pt_tr.b2(zz).squeeze() == BZ_C


@pytest.mark.parametrize('c2', [(ZZ, BZ), BZ_C, None])
def test_pt_tracer_ia_smoke(c2):
    pt_tr = ccl.nl_pt.PTIntrinsicAlignmentTracer((ZZ, BZ),
                                                 c2=c2,
                                                 cdelta=(ZZ, BZ))

    # Test c1 and cdelta do the right thing
    for b in [pt_tr.c1, pt_tr.cdelta]:
        assert b(0.2) == BZ_C

    # Test c2 does the right thing
    if c2 is not None:
        assert pt_tr.c2(0.2) == BZ_C
        zz = np.array([0.2])
        assert pt_tr.c2(zz).squeeze() == BZ_C


def test_pt_tracer_get_bias():
    pt_tr = ccl.nl_pt.PTNumberCountsTracer((ZZ, BZ),
                                           b2=(ZZ, BZ),
                                           bs=(ZZ, BZ))
    b = pt_tr.get_bias('b1', 0.1)
    assert b == BZ_C

    with pytest.raises(KeyError):
        pt_tr.get_bias('b_one', 0.1)


def test_pt_workspace_smoke():
    w = ccl.nl_pt.PTCalculator(log10k_min=-3,
                               log10k_max=1,
                               nk_per_decade=10,
                               pad_factor=2)
    assert len(w.ks) == 40


@pytest.mark.parametrize('options', [['TG', 'TG', False, False, PTC],
                                     ['TG', 'TG', False, False, None],
                                     ['TG', 'TG', False, True, PTC],
                                     ['TG', 'TI', False, False, PTC],
                                     ['TG', 'TM', False, False, PTC],
                                     ['TI', 'TG', False, False, PTC],
                                     ['TI', 'TI', False, False, PTC],
                                     ['TI', 'TI', True, False, PTC],
                                     ['TI', 'TM', False, False, PTC],
                                     ['TM', 'TG', False, False, PTC],
                                     ['TM', 'TI', False, False, PTC],
                                     ['TM', 'TM', False, False, PTC]])
def test_pt_get_pk2d_smoke(options):
    if options[0] == options[1]:
        t2 = None
    else:
        t2 = TRS[options[1]]
    pk = ccl.nl_pt.get_pt_pk2d(COSMO,
                               TRS[options[0]],
                               tracer2=t2,
                               return_ia_bb=options[2],
                               sub_lowk=options[3],
                               ptc=options[4])
    assert isinstance(pk, ccl.Pk2D)


def test_pt_pk2d_bb():
    pee = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TI'], ptc=PTC)
    pbb = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TI'], ptc=PTC,
                                return_ia_bb=True)
    pee2, pbb2 = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TI'], ptc=PTC,
                                       return_ia_ee_and_bb=True)
    assert pee.eval(0.1, 0.9, COSMO) == pee2.eval(0.1, 0.9, COSMO)
    assert pbb.eval(0.1, 0.9, COSMO) == pbb2.eval(0.1, 0.9, COSMO)


@pytest.mark.parametrize('nl', ['nonlinear', 'linear', 'spt'])
def test_pt_get_pk2d_nl(nl):
    pk = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TG'],
                               nonlin_pk_type=nl)
    assert isinstance(pk, ccl.Pk2D)


def test_ptc_raises():
    with pytest.raises(ValueError):
        PTC.update_pk(np.zeros(4))


@pytest.mark.parametrize('typ_nlin,typ_nloc', [('linear', 'nonlinear'),
                                               ('nonlinear', 'linear'),
                                               ('nonlinear', 'spt')])
def test_k2pk_types(typ_nlin, typ_nloc):
    tg = ccl.nl_pt.PTNumberCountsTracer(1., 0., 0., bk2=1.)
    tm = ccl.nl_pt.PTNumberCountsTracer(1., 0., 0.)
    pkmm = ccl.nl_pt.get_pt_pk2d(COSMO, tm, tracer2=tm, ptc=PTC,
                                 nonlin_pk_type=typ_nlin)
    pkmm_t = ccl.nl_pt.get_pt_pk2d(COSMO, tm, tracer2=tm, ptc=PTC,
                                   nonlin_pk_type=typ_nloc)
    pkgg = ccl.nl_pt.get_pt_pk2d(COSMO, tg, tracer2=tg, ptc=PTC,
                                 nonloc_pk_type=typ_nloc)
    ks = np.geomspace(1E-3, 1E1, 128)
    p1 = pkgg.eval(ks, 1., COSMO)
    p2 = pkmm.eval(ks, 1., COSMO)+ks**2*pkmm_t.eval(ks, 1., COSMO)
    assert np.all(np.fabs(p1/p2-1) < 1E-4)


def test_k2pk():
    # Tests the k2 term scaling
    ptc = ccl.nl_pt.PTCalculator(with_NC=True)

    zs = np.array([0., 1.])
    gs4 = ccl.growth_factor(COSMO, 1./(1+zs))**4
    pk_lin_z0 = ccl.linear_matter_power(COSMO, ptc.ks, 1.)
    ptc.update_pk(pk_lin_z0)

    Pd1d1 = np.array([ccl.linear_matter_power(COSMO, ptc.ks, a)
                      for a in 1./(1+zs)]).T
    one = np.ones_like(zs)
    zero = np.zeros_like(zs)
    pmm = ptc.get_pgg(Pd1d1, gs4, one, zero, zero,
                      one, zero, zero, True)
    pmm_b = ptc.get_pgm(Pd1d1, gs4, one, zero, zero)
    pmk = ptc.get_pgg(Pd1d1, gs4, one, zero, zero,
                      one, zero, zero, True, bk21=one)
    pmk_b = ptc.get_pgm(Pd1d1, gs4, one, zero, zero, bk2=one)
    pkk = ptc.get_pgg(Pd1d1, gs4, one, zero, zero, one, zero, zero, True,
                      bk21=one, bk22=one)
    ks = ptc.ks[:, None]
    assert np.all(np.fabs(pmm/Pd1d1-1) < 1E-10)
    assert np.all(np.fabs(pmm_b/Pd1d1-1) < 1E-10)
    assert np.all(np.fabs(pmk/(pmm*(1+0.5*ks**2))-1) < 1E-10)
    assert np.all(np.fabs(pmk_b/(pmm*(1+0.5*ks**2))-1) < 1E-10)
    assert np.all(np.fabs(pkk/(pmm*(1+ks**2))-1) < 1E-10)


def test_pk_cutoff():
    # Tests the exponential cutoff
    ptc1 = ccl.nl_pt.PTCalculator(with_NC=True)
    ptc2 = ccl.nl_pt.PTCalculator(with_NC=True,
                                  k_cutoff=10.,
                                  n_exp_cutoff=2.)
    zs = np.array([0., 1.])
    gs4 = ccl.growth_factor(COSMO, 1./(1+zs))**4
    pk_lin_z0 = ccl.linear_matter_power(COSMO, ptc1.ks, 1.)
    ptc1.update_pk(pk_lin_z0)
    ptc2.update_pk(pk_lin_z0)

    Pd1d1 = np.array([ccl.linear_matter_power(COSMO, ptc1.ks, a)
                      for a in 1./(1+zs)]).T
    one = np.ones_like(zs)
    zero = np.zeros_like(zs)
    p1 = ptc1.get_pgg(Pd1d1, gs4, one, zero, zero,
                      one, zero, zero, True).T
    p2 = ptc2.get_pgg(Pd1d1, gs4, one, zero, zero,
                      one, zero, zero, True).T
    expcut = np.exp(-(ptc1.ks/10.)**2)
    assert np.all(np.fabs(p1*expcut/p2-1) < 1E-10)


def test_pt_get_pk2d_raises():
    # Wrong tracer type 2
    with pytest.raises(TypeError):
        ccl.nl_pt.get_pt_pk2d(COSMO,
                              TRS['TG'],
                              tracer2=3,
                              ptc=PTC)
    # Wrong tracer type 1
    with pytest.raises(TypeError):
        ccl.nl_pt.get_pt_pk2d(COSMO,
                              3,
                              tracer2=TRS['TG'],
                              ptc=PTC)
    # Wrong calculator type
    with pytest.raises(TypeError):
        ccl.nl_pt.get_pt_pk2d(COSMO,
                              TRS['TG'],
                              ptc=3)

    # Incomplete calculator
    ptc_empty = ccl.nl_pt.PTCalculator(with_NC=False,
                                       with_IA=False,
                                       with_dd=False)
    for t in ['TG', 'TI', 'TM']:
        with pytest.raises(ValueError):
            ccl.nl_pt.get_pt_pk2d(COSMO, TRS[t],
                                  nonlin_pk_type='spt',
                                  ptc=ptc_empty)

    # Wrong non-linear Pk
    with pytest.raises(NotImplementedError):
        ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TM'],
                              nonlin_pk_type='halofat')

    # Wrong tracer types
    tdum = ccl.nl_pt.PTMatterTracer()
    tdum.type = 'A'
    for t in ['TG', 'TI', 'TM']:
        with pytest.raises(NotImplementedError):
            ccl.nl_pt.get_pt_pk2d(COSMO, TRS[t],
                                  tracer2=tdum, ptc=PTC)
    with pytest.raises(NotImplementedError):
        ccl.nl_pt.get_pt_pk2d(COSMO, tdum,
                              tracer2=TRS['TM'], ptc=PTC)


def test_translate_IA_norm():
    # test that it works with scalar a, vector z
    c_1, c_d, c_2 = ccl.nl_pt.translate_IA_norm(COSMO, ZZ, a1=a_1, a1delta=a_d,
                                                a2=a_2, Om_m2_for_c2=False)
    assert c_1.all() == c_1_t.all()
    assert c_d.all() == c_d_t.all()
    assert c_2.all() == c_2_t_des.all()

    c_1, c_d, c_2 = ccl.nl_pt.translate_IA_norm(COSMO, ZZ, a1=a_1, a1delta=a_d,
                                                a2=a_2, Om_m2_for_c2=True)
    assert c_2.all() == c_2_t.all()

    # test that it works with scalar a, scalar z
    c_1, c_d, c_2 = ccl.nl_pt.translate_IA_norm(COSMO, ZZ_1, a1=a_1,
                                                Om_m2_for_c2=False)
    assert c_1 == c_1_t_1

    # test that it works with vector a, vector z
    c_1, c_d, c_2 = ccl.nl_pt.translate_IA_norm(COSMO, ZZ, a1=a_1_v,
                                                Om_m2_for_c2=False)
    assert c_1.all() == c_1_t.all()


def test_translate_IA_norm_raises():
    # Should raise error with 2d input a or z
    z_wrong = np.ones((2, 3))
    a_wrong = np.ones((2, 3))
    with pytest.raises(ValueError):
        c_1, c_d, c_2 = ccl.nl_pt.translate_IA_norm(COSMO, ZZ, a1=a_wrong,
                                                    Om_m2_for_c2=False)
    with pytest.raises(ValueError):
        c_1, c_d, c_2 = ccl.nl_pt.translate_IA_norm(COSMO, z_wrong, a1=a_1_v,
                                                    Om_m2_for_c2=False)

    # Should raise error if len(a) != len(z)
    NZ2 = 129
    ZZ2 = np.linspace(0., 1., NZ2)
    with pytest.raises(ValueError):
        c_1, c_d, c_2 = ccl.nl_pt.translate_IA_norm(COSMO, ZZ2, a1=a_1_v,
                                                    Om_m2_for_c2=False)


def test_return_ptc():
    # if no ptc is input, check that returned pk and ptc objects have
    # the correct properties.
    pk, ptc1 = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TG'], return_ptc=True)
    assert isinstance(pk, ccl.Pk2D)
    assert isinstance(ptc1, ccl.nl_pt.power.PTCalculator)
    # same test with EE/BB output
    pee2, pbb2, ptc2 = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TI'],
                                             return_ia_ee_and_bb=True,
                                             return_ptc=True)
    assert isinstance(pee2, ccl.Pk2D)
    assert isinstance(pbb2, ccl.Pk2D)
    assert isinstance(ptc2, ccl.nl_pt.power.PTCalculator)
    # check that returned ptc matches input ptc.
    pk_2, ptc1_2 = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TG'], ptc=PTC,
                                         return_ptc=True)
    pee2_2, pbb2_2, ptc2_2 = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TI'], ptc=PTC,
                                                   return_ia_ee_and_bb=True,
                                                   return_ptc=True)
    assert ptc1_2 is PTC
    assert ptc2_2 is PTC
    # check that the result outputs are the same
    # for the internally initialized ptc.
    assert np.allclose(pk_2.eval(ks, 1., COSMO), pk.eval(ks, 1., COSMO))
    assert np.allclose(pee2_2.eval(ks, 1., COSMO), pee2.eval(ks, 1., COSMO))
    assert np.allclose(pbb2_2.eval(ks, 1., COSMO), pbb2.eval(ks, 1., COSMO))


def test_pt_no_ptc_update():
    pee1 = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TI'], ptc=PTC)
    pee1_no_update = ccl.nl_pt.get_pt_pk2d(COSMO, TRS['TI'], ptc=PTC,
                                           update_ptc=False)

    COSMO2 = ccl.Cosmology(
        Omega_c=COSMO["Omega_c"], Omega_b=COSMO["Omega_b"], h=COSMO["h"],
        sigma8=COSMO["sigma8"]+0.05, n_s=COSMO["n_s"],
        transfer_function='bbks', matter_power_spectrum='linear')

    pee2_no_update = ccl.nl_pt.get_pt_pk2d(COSMO2, TRS['TI'], ptc=PTC,
                                           update_ptc=False)
    pee2 = ccl.nl_pt.get_pt_pk2d(COSMO2, TRS['TI'], ptc=PTC)

    assert pee1.eval(0.1, 0.9, COSMO) == pee1_no_update.eval(0.1, 0.9, COSMO)
    assert pee2.eval(0.1, 0.9, COSMO) != pee2_no_update.eval(0.1, 0.9, COSMO2)



from .. import ccllib as lib
from ..core import check
from ..background import omega_x
from ..parameters import physical_constants
from .massdef import MassDef, MassDef200m
import numpy as np
import functools


class MassFunc(object):
    """ This class enables the calculation of halo mass functions.
    We currently assume that all mass functions can be written as

    .. math::
        \\frac{dn}{d\\log_{10}M} = f(\\sigma_M)\\,\\frac{\\rho_M}{M}\\,
        \\frac{d\\log \\sigma_M}{d\\log_{10} M}

    where :math:`\\sigma_M^2` is the overdensity variance on spheres with a
    radius given by the Lagrangian radius for mass M.
    All sub-classes implementing specific mass function parametrizations
    can therefore be simply created by replacing this class'
    `_get_fsigma` method.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object that fixes
            the mass definition used by this mass function
            parametrization.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = 'default'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        # Initialize sigma(M) splines if needed
        cosmo.compute_sigma()
        self.mass_def_strict = mass_def_strict
        # Check if mass function was provided and check that it's
        # sensible.
        if mass_def is not None:
            if self._check_mdef(mass_def):
                raise ValueError("Mass function " + self.name +
                                 " is not compatible with mass definition" +
                                 " Delta = %s, " % (mass_def.Delta) +
                                 " rho = " + mass_def.rho_type)
            self.mdef = mass_def
        else:
            self._default_mdef()
        self._setup(cosmo)

    def _default_mdef(self):
        """ Assigns a default mass definition for this object if
        none is passed at initialization.
        """
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        """ Use this function to initialize any internal attributes
        of this object. This function is called at the very end of the
        constructor call.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        """
        pass

    def _check_mdef_strict(self, mdef):
        return False

    def _check_mdef(self, mdef):
        """ Return False if the input mass definition agrees with
        the definitions for which this mass function parametrization
        works. True otherwise. This function gets called at the
        start of the constructor call.

        Args:
            mdef (:class:`~pyccl.halos.massdef.MassDef`):
                a mass definition object.

        Returns:
            bool: True if the mass definition is not compatible with \
                this mass function parametrization. False otherwise.
        """
        if self.mass_def_strict:
            return self._check_mdef_strict(mdef)
        return False

    def _get_consistent_mass(self, cosmo, M, a, mdef_other):
        """ Transform a halo mass with a given mass definition into
        the corresponding mass definition that was used to initialize
        this object.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            M (float or array_like): halo mass in units of M_sun.
            a (float): scale factor.
            mdef_other (:class:`~pyccl.halos.massdef.MassDef`):
                a mass definition object.

        Returns:
            float or array_like: mass according to this object's \
                mass definition.
        """
        if mdef_other is not None:
            M_use = mdef_other.translate_mass(cosmo, M, a, self.mdef)
        else:
            M_use = M
        return np.log10(M_use)

    def _get_Delta_m(self, cosmo, a):
        """ For SO-based mass definitions, this returns the corresponding
        value of Delta for a rho_matter-based definition. This is useful
        mostly for the Tinker mass functions, which are defined for any
        SO mass in general, but explicitly only for Delta_matter.
        """
        delta = self.mdef.get_Delta(cosmo, a)
        if self.mdef.rho_type == 'matter':
            return delta
        else:
            om_this = omega_x(cosmo, a, self.mdef.rho_type)
            om_matt = omega_x(cosmo, a, 'matter')
            return delta * om_this / om_matt

    def get_mass_function(self, cosmo, M, a, mdef_other=None):
        """ Returns the mass function for input parameters.

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            M (float or array_like): halo mass in units of M_sun.
            a (float): scale factor.
            mdef_other (:class:`~pyccl.halos.massdef.MassDef`):
                the mass definition object that defines M.

        Returns:
            float or array_like: mass function \
                :math:`dn/d\\log_{10}M` in units of Mpc^-3 (comoving).
        """
        M_use = np.atleast_1d(M)
        logM = self._get_consistent_mass(cosmo, M_use,
                                         a, mdef_other)

        # sigma(M)
        status = 0
        sigM, status = lib.sigM_vec(cosmo.cosmo, a, logM,
                                    len(logM), status)
        check(status, cosmo=cosmo)
        # dlogsigma(M)/dlog10(M)
        dlns_dlogM, status = lib.dlnsigM_dlogM_vec(cosmo.cosmo, a, logM,
                                                   len(logM), status)
        check(status, cosmo=cosmo)

        rho = (physical_constants.RHO_CRITICAL *
               cosmo['Omega_m'] * cosmo['h']**2)
        f = self._get_fsigma(cosmo, sigM, a, 2.302585092994046 * logM)
        mf = f * rho * dlns_dlogM / M_use

        if np.ndim(M) == 0:
            mf = mf[0]
        return mf

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        """ Get the :math:`f(\\sigma_M)` function for this mass function
        object (see description of this class for details).

        Args:
            cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
            sigM (float or array_like): standard deviation in the
                overdensity field on the scale of this halo.
            a (float): scale factor.
            lnM (float or array_like): natural logarithm of the
                halo mass in units of M_sun (provided in addition
                to sigM for convenience in some mass function
                parametrizations).

        Returns:
            float or array_like: :math:`f(\\sigma_M)` function.
        """
        raise NotImplementedError("Use one of the non-default "
                                  "MassFunction classes")

    @classmethod
    def from_name(cls, name):
        """ Returns mass function subclass from name string

        Args:
            name (string): a mass function name

        Returns:
            MassFunc subclass corresponding to the input name.
        """
        mass_functions = {c.name: c for c in cls.__subclasses__()}
        if name in mass_functions:
            return mass_functions[name]
        else:
            raise ValueError(f"Mass function {name} not implemented.")


class MassFuncPress74(MassFunc):
    """ Implements mass function described in 1974ApJ...187..425P.
    This parametrization is only valid for 'fof' masses.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts FoF masses only.
            If `None`, FoF masses will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = 'Press74'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        super(MassFuncPress74, self).__init__(cosmo,
                                              mass_def,
                                              mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        self.norm = np.sqrt(2/np.pi)

    def _check_mdef_strict(self, mdef):
        if mdef.Delta != 'fof':
            return True
        return False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        delta_c = 1.68647

        nu = delta_c/sigM
        return self.norm * nu * np.exp(-0.5 * nu**2)


class MassFuncSheth99(MassFunc):
    """ Implements mass function described in arXiv:astro-ph/9901122
    This parametrization is only valid for 'fof' masses.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts FoF masses only.
            If `None`, FoF masses will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
        use_delta_c_fit (bool): if True, use delta_crit given by
            the fit of Nakamura & Suto 1997. Otherwise use
            delta_crit = 1.68647.
    """
    name = 'Sheth99'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True,
                 use_delta_c_fit=False):
        self.use_delta_c_fit = use_delta_c_fit
        super(MassFuncSheth99, self).__init__(cosmo,
                                              mass_def,
                                              mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        self.A = 0.21615998645
        self.p = 0.3
        self.a = 0.707

    def _check_mdef_strict(self, mdef):
        if mdef.Delta != 'fof':
            return True

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        if self.use_delta_c_fit:
            status = 0
            delta_c, status = lib.dc_NakamuraSuto(cosmo.cosmo, a, status)
            check(status, cosmo=cosmo)
        else:
            delta_c = 1.68647

        nu = delta_c / sigM
        return nu * self.A * (1. + (self.a * nu**2)**(-self.p)) * \
            np.exp(-self.a * nu**2/2.)


class MassFuncJenkins01(MassFunc):
    """ Implements mass function described in astro-ph/0005260.
    This parametrization is only valid for 'fof' masses.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts FoF masses only.
            If `None`, FoF masses will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = 'Jenkins01'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        super(MassFuncJenkins01, self).__init__(cosmo,
                                                mass_def=mass_def,
                                                mass_def_strict=True)

    def _default_mdef(self):
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        self.A = 0.315
        self.b = 0.61
        self.q = 3.8

    def _check_mdef_strict(self, mdef):
        if mdef.Delta != 'fof':
            return True
        return False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        return self.A * np.exp(-np.fabs(-np.log(sigM) + self.b)**self.q)


class MassFuncTinker08(MassFunc):
    """ Implements mass function described in arXiv:0803.2706.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts SO masses with
            200 < Delta < 3200 with respect to the matter density.
            If `None`, Delta = 200 (matter) will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = 'Tinker08'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        super(MassFuncTinker08, self).__init__(cosmo,
                                               mass_def,
                                               mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef200m()

    def _pd(self, ld):
        return 10.**(-(0.75/(ld - 1.8750612633))**1.2)

    def _setup(self, cosmo):
        from scipy.interpolate import interp1d

        delta = np.array([200.0, 300.0, 400.0, 600.0, 800.0,
                          1200.0, 1600.0, 2400.0, 3200.0])
        alpha = np.array([0.186, 0.200, 0.212, 0.218, 0.248,
                          0.255, 0.260, 0.260, 0.260])
        beta = np.array([1.47, 1.52, 1.56, 1.61, 1.87,
                         2.13, 2.30, 2.53, 2.66])
        gamma = np.array([2.57, 2.25, 2.05, 1.87, 1.59,
                          1.51, 1.46, 1.44, 1.41])
        phi = np.array([1.19, 1.27, 1.34, 1.45, 1.58,
                        1.80, 1.97, 2.24, 2.44])
        ldelta = np.log10(delta)
        self.pA0 = interp1d(ldelta, alpha)
        self.pa0 = interp1d(ldelta, beta)
        self.pb0 = interp1d(ldelta, gamma)
        self.pc = interp1d(ldelta, phi)

    def _check_mdef_strict(self, mdef):
        if mdef.Delta == 'fof':
            return True
        return False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        ld = np.log10(self._get_Delta_m(cosmo, a))
        pA = self.pA0(ld) * a**0.14
        pa = self.pa0(ld) * a**0.06
        pb = self.pb0(ld) * a**self._pd(ld)
        return pA * ((pb / sigM)**pa + 1) * np.exp(-self.pc(ld)/sigM**2)


class MassFuncDespali16(MassFunc):
    """ Implements mass function described in arXiv:1507.05627.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts any SO masses.
            If `None`, Delta = 200 (matter) will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = 'Despali16'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True,
                 ellipsoidal=False):
        super(MassFuncDespali16, self).__init__(cosmo,
                                                mass_def,
                                                mass_def_strict)
        self.ellipsoidal = ellipsoidal

    def _default_mdef(self):
        self.mdef = MassDef200m()

    def _setup(self, cosmo):
        pass

    def _check_mdef_strict(self, mdef):
        if mdef.Delta == 'fof':
            return True
        return False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        status = 0
        delta_c, status = lib.dc_NakamuraSuto(cosmo.cosmo, a, status)
        check(status, cosmo=cosmo)

        Dv, status = lib.Dv_BryanNorman(cosmo.cosmo, a, status)
        check(status, cosmo=cosmo)

        x = np.log10(self.mdef.get_Delta(cosmo, a) *
                     omega_x(cosmo, a, self.mdef.rho_type) / Dv)

        if self.ellipsoidal:
            A = -0.1768 * x + 0.3953
            a = 0.3268 * x**2 + 0.2125 * x + 0.7057
            p = -0.04570 * x**2 + 0.1937 * x + 0.2206
        else:
            A = -0.1362 * x + 0.3292
            a = 0.4332 * x**2 + 0.2263 * x + 0.7665
            p = -0.1151 * x**2 + 0.2554 * x + 0.2488

        nu = delta_c/sigM
        nu_p = a * nu**2

        return 2.0 * A * np.sqrt(nu_p / 2.0 / np.pi) * \
            np.exp(-0.5 * nu_p) * (1.0 + nu_p**-p)


class MassFuncTinker10(MassFunc):
    """ Implements mass function described in arXiv:1001.3162.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts SO masses with
            200 < Delta < 3200 with respect to the matter density.
            If `None`, Delta = 200 (matter) will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
        norm_all_z (bool): should we normalize the mass function
            at z=0 or at all z?
    """
    name = 'Tinker10'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True,
                 norm_all_z=False):
        self.norm_all_z = norm_all_z
        super(MassFuncTinker10, self).__init__(cosmo,
                                               mass_def,
                                               mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef200m()

    def _setup(self, cosmo):
        from scipy.interpolate import interp1d

        delta = np.array([200.0, 300.0, 400.0, 600.0, 800.0,
                          1200.0, 1600.0, 2400.0, 3200.0])
        alpha = np.array([0.368, 0.363, 0.385, 0.389, 0.393,
                          0.365, 0.379, 0.355, 0.327])
        beta = np.array([0.589, 0.585, 0.544, 0.543, 0.564,
                         0.623, 0.637, 0.673, 0.702])
        gamma = np.array([0.864, 0.922, 0.987, 1.09, 1.20,
                          1.34, 1.50, 1.68, 1.81])
        phi = np.array([-0.729, -0.789, -0.910, -1.05, -1.20,
                        -1.26, -1.45, -1.50, -1.49])
        eta = np.array([-0.243, -0.261, -0.261, -0.273, -0.278,
                        -0.301, -0.301, -0.319, -0.336])

        ldelta = np.log10(delta)
        self.pA0 = interp1d(ldelta, alpha)
        self.pa0 = interp1d(ldelta, eta)
        self.pb0 = interp1d(ldelta, beta)
        self.pc0 = interp1d(ldelta, gamma)
        self.pd0 = interp1d(ldelta, phi)
        if self.norm_all_z:
            p = np.array([-0.158, -0.195, -0.213, -0.254, -0.281,
                          -0.349, -0.367, -0.435, -0.504])
            q = np.array([0.0128, 0.0128, 0.0143, 0.0154, 0.0172,
                          0.0174, 0.0199, 0.0203, 0.0205])
            self.pp0 = interp1d(ldelta, p)
            self.pq0 = interp1d(ldelta, q)

    def _check_mdef_strict(self, mdef):
        if mdef.Delta == 'fof':
            return True
        return False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        ld = np.log10(self._get_Delta_m(cosmo, a))
        nu = 1.686 / sigM
        # redshift evolution only up to z=3
        a = np.clip(a, 0.25, 1)
        pa = self.pa0(ld) * a**(-0.27)
        pb = self.pb0(ld) * a**(-0.20)
        pc = self.pc0(ld) * a**0.01
        pd = self.pd0(ld) * a**0.08
        pA0 = self.pA0(ld)
        if self.norm_all_z:
            z = 1./a - 1
            pp = self.pp0(ld)
            pq = self.pq0(ld)
            pA0 *= np.exp(z*(pp+pq*z))
        return nu * pA0 * (1 + (pb * nu)**(-2 * pd)) * \
            nu**(2 * pa) * np.exp(-0.5 * pc * nu**2)


class MassFuncBocquet16(MassFunc):
    """ Implements mass function described in arXiv:1502.07357.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts SO masses with
            Delta = 200 (matter, critical) and 500 (critical).
            If `None`, Delta = 200 (matter) will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
        hydro (bool): if `False`, use the parametrization found
            using dark-matter-only simulations. Otherwise, include
            baryonic effects (default).
    """
    name = 'Bocquet16'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True,
                 hydro=True):
        self.hydro = hydro
        super(MassFuncBocquet16, self).__init__(cosmo,
                                                mass_def,
                                                mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef200m()

    def _setup(self, cosmo):
        if int(self.mdef.Delta) == 200:
            if self.mdef.rho_type == 'matter':
                self.mdef_type = '200m'
            elif self.mdef.rho_type == 'critical':
                self.mdef_type = '200c'
        elif int(self.mdef.Delta) == 500:
            if self.mdef.rho_type == 'critical':
                self.mdef_type = '500c'
        if self.mdef_type == '200m':
            if self.hydro:
                self.A0 = 0.228
                self.a0 = 2.15
                self.b0 = 1.69
                self.c0 = 1.30
                self.Az = 0.285
                self.az = -0.058
                self.bz = -0.366
                self.cz = -0.045
            else:
                self.A0 = 0.175
                self.a0 = 1.53
                self.b0 = 2.55
                self.c0 = 1.19
                self.Az = -0.012
                self.az = -0.040
                self.bz = -0.194
                self.cz = -0.021
        elif self.mdef_type == '200c':
            if self.hydro:
                self.A0 = 0.202
                self.a0 = 2.21
                self.b0 = 2.00
                self.c0 = 1.57
                self.Az = 1.147
                self.az = 0.375
                self.bz = -1.074
                self.cz = -0.196
            else:
                self.A0 = 0.222
                self.a0 = 1.71
                self.b0 = 2.24
                self.c0 = 1.46
                self.Az = 0.269
                self.az = 0.321
                self.bz = -0.621
                self.cz = -0.153
        elif self.mdef_type == '500c':
            if self.hydro:
                self.A0 = 0.180
                self.a0 = 2.29
                self.b0 = 2.44
                self.c0 = 1.97
                self.Az = 1.088
                self.az = 0.150
                self.bz = -1.008
                self.cz = -0.322
            else:
                self.A0 = 0.241
                self.a0 = 2.18
                self.b0 = 2.35
                self.c0 = 2.02
                self.Az = 0.370
                self.az = 0.251
                self.bz = -0.698
                self.cz = -0.310

    def _check_mdef_strict(self, mdef):
        if isinstance(mdef.Delta, str):
            return True
        elif int(mdef.Delta) == 200:
            if (mdef.rho_type != 'matter') and \
               (mdef.rho_type != 'critical'):
                return True
        elif int(mdef.Delta) == 500:
            if mdef.rho_type != 'critical':
                return True
        else:
            return True
        return False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        zp1 = 1./a
        AA = self.A0 * zp1**self.Az
        aa = self.a0 * zp1**self.az
        bb = self.b0 * zp1**self.bz
        cc = self.c0 * zp1**self.cz

        f = AA * ((sigM / bb)**-aa + 1.0) * np.exp(-cc / sigM**2)

        if self.mdef_type == '200c':
            z = 1./a-1
            Omega_m = omega_x(cosmo, a, "matter")
            gamma0 = 3.54E-2 + Omega_m**0.09
            gamma1 = 4.56E-2 + 2.68E-2 / Omega_m
            gamma2 = 0.721 + 3.50E-2 / Omega_m
            gamma3 = 0.628 + 0.164 / Omega_m
            delta0 = -1.67E-2 + 2.18E-2 * Omega_m
            delta1 = 6.52E-3 - 6.86E-3 * Omega_m
            gamma = gamma0 + gamma1 * np.exp(-((gamma2 - z) / gamma3)**2)
            delta = delta0 + delta1 * z
            M200c_M200m = gamma + delta * lnM
            f *= M200c_M200m
        elif self.mdef_type == '500c':
            z = 1./a-1
            Omega_m = omega_x(cosmo, a, "matter")
            alpha0 = 0.880 + 0.329 * Omega_m
            alpha1 = 1.00 + 4.31E-2 / Omega_m
            alpha2 = -0.365 + 0.254 / Omega_m
            alpha = alpha0 * (alpha1 * z + alpha2) / (z + alpha2)
            beta = -1.7E-2 + 3.74E-3 * Omega_m
            M500c_M200m = alpha + beta * lnM
            f *= M500c_M200m
        return f


class MassFuncWatson13(MassFunc):
    """ Implements mass function described in arXiv:1212.0095.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts fof and any SO masses.
            If `None`, Delta = 200 (matter) will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = 'Watson13'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        super(MassFuncWatson13, self).__init__(cosmo,
                                               mass_def,
                                               mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef200m()

    def _setup(self, cosmo):
        self.is_fof = self.mdef.Delta == 'fof'

    def _check_mdef_strict(self, mdef):
        if mdef.Delta == 'vir':
            return True
        return False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        if self.is_fof:
            pA = 0.282
            pa = 2.163
            pb = 1.406
            pc = 1.210
            return pA * ((pb / sigM)**pa + 1.) * np.exp(-pc / sigM**2)
        else:
            om = omega_x(cosmo, a, "matter")
            Delta_178 = self.mdef.Delta / 178.0

            if a == 1.0:
                pA = 0.194
                pa = 1.805
                pb = 2.267
                pc = 1.287
            elif a < 0.14285714285714285:  # z>6
                pA = 0.563
                pa = 3.810
                pb = 0.874
                pc = 1.453
            else:
                pA = om * (1.097 * a**3.216 + 0.074)
                pa = om * (5.907 * a**3.058 + 2.349)
                pb = om * (3.136 * a**3.599 + 2.344)
                pc = 1.318

            f_178 = pA * ((pb / sigM)**pa + 1.) * np.exp(-pc / sigM**2)
            C = np.exp(0.023 * (Delta_178 - 1.0))
            d = -0.456 * om - 0.139
            Gamma = (C * Delta_178**d *
                     np.exp(0.072 * (1.0 - Delta_178) / sigM**2.130))
            return f_178 * Gamma


class MassFuncAngulo12(MassFunc):
    """ Implements mass function described in arXiv:1203.3216.
    This parametrization is only valid for 'fof' masses.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        mass_def (:class:`~pyccl.halos.massdef.MassDef`):
            a mass definition object.
            this parametrization accepts FoF masses only.
            If `None`, FoF masses will be used.
        mass_def_strict (bool): if False, consistency of the mass
            definition will be ignored.
    """
    name = 'Angulo12'

    def __init__(self, cosmo, mass_def=None, mass_def_strict=True):
        super(MassFuncAngulo12, self).__init__(cosmo,
                                               mass_def,
                                               mass_def_strict)

    def _default_mdef(self):
        self.mdef = MassDef('fof', 'matter')

    def _setup(self, cosmo):
        self.A = 0.201
        self.a = 2.08
        self.b = 1.7
        self.c = 1.172

    def _check_mdef_strict(self, mdef):
        if mdef.Delta != 'fof':
            return True
        return False

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        return self.A * ((self.a / sigM)**self.b + 1.) * \
            np.exp(-self.c / sigM**2)


@functools.wraps(MassFunc.from_name)
def mass_function_from_name(name):
    return MassFunc.from_name(name)



import numpy as np

from . import ccllib as lib
from .pyutils import check, integ_types, _check_array_params
from .background import comoving_radial_distance, comoving_angular_distance
from .tk3d import Tk3D
from .pk2d import parse_pk2d

# Define symbolic 'None' type for arrays, to allow proper handling by swig
# wrapper
NoneArr = np.array([])


def angular_cl_cov_cNG(cosmo, cltracer1, cltracer2, ell, tkka, fsky=1.,
                       cltracer3=None, cltracer4=None, ell2=None,
                       integration_method='qag_quad'):
    """Calculate the connected non-Gaussian covariance for a pair of
    power spectra :math:`C_{\\ell_1}^{ab}` and :math:`C_{\\ell_2}^{cd}`,
    between two pairs of tracers (:math:`(a,b)` and :math:`(c,d)`).

    Specifically, it computes:

    .. math::
        {\\rm Cov}_{\\rm cNG}(\\ell_1,\\ell_2)=
        \\int \\frac{d\\chi}{\\chi^6}
        \\tilde{\\Delta}^a_{\\ell_1}(\\chi)
        \\tilde{\\Delta}^b_{\\ell_1}(\\chi)
        \\tilde{\\Delta}^c_{\\ell_2}(\\chi)
        \\tilde{\\Delta}^d_{\\ell_2}(\\chi)\\,
        \\bar{T}_{abcd}\\left[\\frac{\\ell_1+1/2}{\\chi},
                              \\frac{\\ell_2+1/2}{\\chi}, a(\\chi)\\right]

    where :math:`\\Delta^x_\\ell(\\chi)` is the transfer function for tracer
    :math:`x` (see Eq. 39 in the CCL note), and
    :math:`\\bar{T}_{abcd}(k_1,k_2,a)` is the isotropized connected
    trispectrum of the four tracers (see the documentation of the
    :class:`~pyccl.tk3d.Tk3D` class for details).

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        cltracer1 (:class:`~pyccl.tracers.Tracer`): a `Tracer` object,
            of any kind.
        cltracer2 (:class:`~pyccl.tracers.Tracer`): a second `Tracer` object,
            of any kind.
        ell (float or array_like): Angular wavenumber(s) at which to evaluate
            the first dimension of the angular power spectrum covariance.
        tkka (:class:`~pyccl.tk3d.Tk3D` or None): 3D connected trispectrum.
        fsky (float): sky fraction.
        cltracer3 (:class:`~pyccl.tracers.Tracer`): a `Tracer` object,
            of any kind. If `None`, `cltracer1` will be used instead.
        cltracer4 (:class:`~pyccl.tracers.Tracer`): a `Tracer` object,
            of any kind. If `None`, `cltracer1` will be used instead.
        ell2 (float or array_like): Angular wavenumber(s) at which to evaluate
            the second dimension of the angular power spectrum covariance. If
            `None`, `ell` will be used instead.
        integration_method (string) : integration method to be used
            for the Limber integrals. Possibilities: 'qag_quad' (GSL's `qag`
            method backed up by `quad` when it fails) and 'spline' (the
            integrand is splined and then integrated analytically).

    Returns:
        float or array_like: 2D array containing the connected non-Gaussian \
            Angular power spectrum covariance \
            :math:`Cov_{\\rm cNG}(\\ell_1,\\ell_2)`, for the \
            four input tracers, as a function of :math:`\\ell_1` and \
            :math:`\\ell_2`. The ordering is such that \
            `out[i2, i1] = Cov(ell2[i2], ell[i1])`.
    """
    if integration_method not in ['qag_quad', 'spline']:
        raise ValueError("Integration method %s not supported" %
                         integration_method)

    # we need the distances for the integrals
    cosmo.compute_distances()

    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    if isinstance(tkka, Tk3D):
        tsp = tkka.tsp
    else:
        raise ValueError("tkka must be a pyccl.Tk3D")

    # Create tracer colections
    status = 0
    clt1, status = lib.cl_tracer_collection_t_new(status)
    for t in cltracer1._trc:
        status = lib.add_cl_tracer_to_collection(clt1, t, status)
    clt2, status = lib.cl_tracer_collection_t_new(status)
    for t in cltracer2._trc:
        status = lib.add_cl_tracer_to_collection(clt2, t, status)
    if cltracer3 is None:
        clt3 = clt1
    else:
        clt3, status = lib.cl_tracer_collection_t_new(status)
        for t in cltracer3._trc:
            status = lib.add_cl_tracer_to_collection(clt3, t, status)
    if cltracer4 is None:
        clt4 = clt2
    else:
        clt4, status = lib.cl_tracer_collection_t_new(status)
        for t in cltracer4._trc:
            status = lib.add_cl_tracer_to_collection(clt4, t, status)

    ell1_use = np.atleast_1d(ell)
    if ell2 is None:
        ell2 = ell
    ell2_use = np.atleast_1d(ell2)

    cov, status = lib.angular_cov_vec(
        cosmo, clt1, clt2, clt3, clt4, tsp,
        ell1_use, ell2_use, integ_types[integration_method],
        6, 1./(4*np.pi*fsky), ell1_use.size*ell2_use.size, status)

    cov = cov.reshape([ell2_use.size, ell1_use.size])
    if np.ndim(ell2) == 0:
        cov = np.squeeze(cov, axis=0)
    if np.ndim(ell) == 0:
        cov = np.squeeze(cov, axis=-1)

    # Free up tracer collections
    lib.cl_tracer_collection_t_free(clt1)
    lib.cl_tracer_collection_t_free(clt2)
    if cltracer3 is not None:
        lib.cl_tracer_collection_t_free(clt3)
    if cltracer4 is not None:
        lib.cl_tracer_collection_t_free(clt4)

    check(status, cosmo=cosmo_in)
    return cov


def sigma2_B_disc(cosmo, a=None, fsky=1., p_of_k_a=None):
    """Returns the variance of the projected linear density field
    over a circular disc covering a sky fraction `fsky` as a function
    of scale factor. This is given by

    .. math::
        \\sigma^2_B(z) = \\int_0^\\infty \\frac{k\\,dk}{2\\pi}
            P_L(k,z)\\,\\left[\\frac{2J_1(k R(z))}{k R(z)}\\right]^2,

    where :math:`R(z)` is the corresponding radial aperture as a
    function of redshift. This quantity is used to compute the
    super-sample covariance.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): a Cosmology object.
        a_arr (float, array_like or `None`): an array of scale factor
            values at which to evaluate the projected variance. If
            `None`, a default sampling will be used.
        fsky (float): sky fraction.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, str, or `None`): Linear
            power spectrum to use. Defaults to `None`, in which case the
            internal linear power spectrum from `cosmo` is used.

    Returns:
        float or array_like: values of the projected variance.
    """
    status = 0
    if a is None:
        na = lib.get_pk_spline_na(cosmo.cosmo)
        a_arr, status = lib.get_pk_spline_a(cosmo.cosmo, na, status)
        check(status, cosmo=cosmo)
    else:
        a_arr = np.atleast_1d(a)
        na = len(a_arr)

    chi_arr = comoving_radial_distance(cosmo, a_arr)
    R_arr = chi_arr * np.arccos(1-2*fsky)
    psp = parse_pk2d(cosmo, p_of_k_a, is_linear=True)

    s2B_arr, status = lib.sigma2b_vec(cosmo.cosmo, a_arr, R_arr, psp,
                                      na, status)
    check(status, cosmo=cosmo)
    if a is None:
        return a_arr, s2B_arr
    else:
        if np.ndim(a) == 0:
            return s2B_arr[0]
        else:
            return s2B_arr


def sigma2_B_from_mask(cosmo, a=None, mask_wl=None, p_of_k_a=None):
    """ Returns the variance of the projected linear density field, given the
        angular power spectrum of the footprint mask and scale factor.
        This is given by

    .. math::
        \\sigma^2_B(z) = \\frac{1}{\\chi^2{z}}\\sum_\\ell
            P_L(\\frac{\\ell+\\frac{1}{2}}{\\chi(z)},z)\\,
            (2\\ell+1)\\sum_m W^A_{\\ell m} {W^B}^*_{\\ell m},

    where :math:`W^A_{\\ell m}` and :math:`W^B_{\\ell m}` are the spherical
    harmonics decomposition of the footprint masks of fields `A` and `B`,
    normalized by their area.

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): a Cosmology object.
        a (float, array_like or `None`): an array of scale factor
            values at which to evaluate the projected variance.
        mask_wl (array_like): Array with the angular power spectrum of the
            masks. The power spectrum should be given at integer multipoles,
            starting at :math:`\\ell=0`. The power spectrum is normalized
            as :math:`(2\\ell+1)\\sum_m W^A_{\\ell m} {W^B}^*_{\\ell m}`. It is
            the responsibility of the user to the provide the mask power out to
            sufficiently high ell for their required precision.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, str, or `None`): Linear
            power spectrum to use. Defaults to `None`, in which case the
            internal linear power spectrum from `cosmo` is used.

    Returns:
        float or array_like: values of the projected variance.
    """
    if p_of_k_a is None:
        cosmo.compute_linear_power()
        p_of_k_a = cosmo.get_linear_power()

    a_arr = np.atleast_1d(a)
    chi = comoving_angular_distance(cosmo, a=a_arr)

    ell = np.arange(mask_wl.size)

    sigma2_B = np.zeros(a_arr.size)
    for i in range(sigma2_B.size):
        if 1-a_arr[i] < 1e-6:
            # For a=1, the integral becomes independent of the footprint in
            # the flat-sky approximation. So we are just using the method
            # for the disc geometry here
            sigma2_B[i] = sigma2_B_disc(cosmo=cosmo, a=a_arr[i],
                                        p_of_k_a=p_of_k_a)
        else:
            k = (ell+0.5)/chi[i]
            pk = p_of_k_a.eval(k, a_arr[i], cosmo)
            # See eq. E.10 of 2007.01844
            sigma2_B[i] = np.sum(pk * mask_wl)/chi[i]**2

    if np.ndim(a) == 0:
        return sigma2_B[0]
    else:
        return sigma2_B


def angular_cl_cov_SSC(cosmo, cltracer1, cltracer2, ell, tkka,
                       sigma2_B=None, fsky=1.,
                       cltracer3=None, cltracer4=None, ell2=None,
                       integration_method='qag_quad'):
    """Calculate the super-sample contribution to the connected
    non-Gaussian covariance for a pair of power spectra
    :math:`C_{\\ell_1}^{ab}` and :math:`C_{\\ell_2}^{cd}`,
    between two pairs of tracers (:math:`(a,b)` and :math:`(c,d)`).

    Specifically, it computes:

    .. math::
        {\\rm Cov}_{\\rm cNG}(\\ell_1,\\ell_2)=
        \\int \\frac{d\\chi}{\\chi^6}
        \\tilde{\\Delta}^a_{\\ell_1}(\\chi)
        \\tilde{\\Delta}^b_{\\ell_1}(\\chi)
        \\tilde{\\Delta}^c_{\\ell_2}(\\chi)
        \\tilde{\\Delta}^d_{\\ell_2}(\\chi)\\,
        \\bar{T}_{abcd}\\left[\\frac{\\ell_1+1/2}{\\chi},
                              \\frac{\\ell_2+1/2}{\\chi}, a(\\chi)\\right]

    where :math:`\\Delta^x_\\ell(\\chi)` is the transfer function for tracer
    :math:`x` (see Eq. 39 in the CCL note), and
    :math:`\\bar{T}_{abcd}(k_1,k_2,a)` is the isotropized connected
    trispectrum of the four tracers (see the documentation of the
    :class:`~pyccl.tk3d.Tk3D` class for details).

    Args:
        cosmo (:class:`~pyccl.core.Cosmology`): A Cosmology object.
        cltracer1 (:class:`~pyccl.tracers.Tracer`): a `Tracer` object,
            of any kind.
        cltracer2 (:class:`~pyccl.tracers.Tracer`): a second `Tracer` object,
            of any kind.
        ell (float or array_like): Angular wavenumber(s) at which to evaluate
            the first dimension of the angular power spectrum covariance.
        tkka (:class:`~pyccl.tk3d.Tk3D` or None): 3D connected trispectrum.
        sigma2_B (tuple of arrays or `None`): A tuple of arrays
            (a, sigma2_B(a)) containing the variance of the projected matter
            overdensity over the footprint as a function of the scale factor.
            If `None`, a compact circular footprint will be assumed covering
            a sky fraction `fsky`.
        fsky (float): sky fraction.
        cltracer3 (:class:`~pyccl.tracers.Tracer`): a `Tracer` object,
            of any kind. If `None`, `cltracer1` will be used instead.
        cltracer4 (:class:`~pyccl.tracers.Tracer`): a `Tracer` object,
            of any kind. If `None`, `cltracer1` will be used instead.
        ell2 (float or array_like): Angular wavenumber(s) at which to evaluate
            the second dimension of the angular power spectrum covariance. If
            `None`, `ell` will be used instead.
        integration_method (string) : integration method to be used
            for the Limber integrals. Possibilities: 'qag_quad' (GSL's `qag`
            method backed up by `quad` when it fails) and 'spline' (the
            integrand is splined and then integrated analytically).

    Returns:
        float or array_like: 2D array containing the super-sample \
            Angular power spectrum covariance \
            :math:`Cov_{\\rm SSC}(\\ell_1,\\ell_2)`, for the \
            four input tracers, as a function of :math:`\\ell_1` and \
            :math:`\\ell_2`. The ordering is such that \
            `out[i2, i1] = Cov(ell2[i2], ell[i1])`.
    """
    if integration_method not in ['qag_quad', 'spline']:
        raise ValueError("Integration method %s not supported" %
                         integration_method)

    # we need the distances for the integrals
    cosmo.compute_distances()

    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    if isinstance(tkka, Tk3D):
        tsp = tkka.tsp
    else:
        raise ValueError("tkka must be a pyccl.Tk3D")

    # Create tracer colections
    status = 0
    clt1, status = lib.cl_tracer_collection_t_new(status)
    for t in cltracer1._trc:
        status = lib.add_cl_tracer_to_collection(clt1, t, status)
    clt2, status = lib.cl_tracer_collection_t_new(status)
    for t in cltracer2._trc:
        status = lib.add_cl_tracer_to_collection(clt2, t, status)
    if cltracer3 is None:
        clt3 = clt1
    else:
        clt3, status = lib.cl_tracer_collection_t_new(status)
        for t in cltracer3._trc:
            status = lib.add_cl_tracer_to_collection(clt3, t, status)
    if cltracer4 is None:
        clt4 = clt2
    else:
        clt4, status = lib.cl_tracer_collection_t_new(status)
        for t in cltracer4._trc:
            status = lib.add_cl_tracer_to_collection(clt4, t, status)

    ell1_use = np.atleast_1d(ell)
    if ell2 is None:
        ell2 = ell
    ell2_use = np.atleast_1d(ell2)

    if sigma2_B is None:
        a_arr, s2b_arr = sigma2_B_disc(cosmo_in, fsky=fsky)
    else:
        a_arr, s2b_arr = _check_array_params(sigma2_B, 'sigma2_B')
    cov, status = lib.angular_cov_ssc_vec(
        cosmo, clt1, clt2, clt3, clt4, tsp, a_arr, s2b_arr,
        ell1_use, ell2_use, integ_types[integration_method],
        4, 1., ell1_use.size*ell2_use.size, status)

    cov = cov.reshape([ell2_use.size, ell1_use.size])
    if np.ndim(ell2) == 0:
        cov = np.squeeze(cov, axis=0)
    if np.ndim(ell) == 0:
        cov = np.squeeze(cov, axis=-1)

    # Free up tracer collections
    lib.cl_tracer_collection_t_free(clt1)
    lib.cl_tracer_collection_t_free(clt2)
    if cltracer3 is not None:
        lib.cl_tracer_collection_t_free(clt3)
    if cltracer4 is not None:
        lib.cl_tracer_collection_t_free(clt4)

    check(status, cosmo=cosmo_in)
    return cov



import os
import numpy as np
import pyccl as ccl

cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, Omega_g=0, Omega_k=0,
                      h=0.7, sigma8=0.8, n_s=0.96, Neff=0, m_nu=0.0,
                      w0=-1, wa=0, T_CMB=2.7255,
                      transfer_function='eisenstein_hu')
dirdat = os.path.join(os.path.dirname(__file__), 'data')
Ms, Rs_200m, Rs_500c, Ms_500c = np.loadtxt(os.path.join(dirdat,
                                                        'mdef_bm.txt'),
                                           unpack=True)
dc = np.loadtxt(os.path.join(dirdat, 'conc_bm.txt'), unpack=True)
Ms = dc[0]
cs_200m_d = dc[1]
cs_200c_d = dc[2]
cs_200m_b = dc[3]
cs_200c_b = dc[4]
cs_vir_k = dc[5]
cs_vir_b = dc[6]
cs_200c_p = dc[7]
cs_200c_di = dc[8]

hmd_vir = ccl.halos.MassDefVir()
hmd_vir_b = ccl.halos.MassDefVir('Bhattacharya13')
hmd_200m = ccl.halos.MassDef200m()
hmd_200m_b = ccl.halos.MassDef200m('Bhattacharya13')
hmd_200c = ccl.halos.MassDef200c()
hmd_200c_b = ccl.halos.MassDef200c('Bhattacharya13')
hmd_200c_p = ccl.halos.MassDef200c('Prada12')
hmd_200c_di = ccl.halos.MassDef200c('Diemer15')
hmd_500c = ccl.halos.MassDef(500, 'critical')


def test_mdef_get_radius():
    Rs_200m_h = hmd_200m.get_radius(cosmo, Ms, 1.)
    Rs_500c_h = hmd_500c.get_radius(cosmo, Ms, 1.)
    assert np.all(np.fabs(Rs_200m_h/Rs_200m-1) < 1E-6)
    assert np.all(np.fabs(Rs_500c_h/Rs_500c-1) < 1E-6)


def test_mdef_get_mass():
    Ms_h = hmd_200m.get_mass(cosmo, Rs_200m, 1.)
    assert np.all(np.fabs(Ms_h/Ms-1) < 1E-6)
    Ms_h = hmd_500c.get_mass(cosmo, Rs_500c, 1.)
    assert np.all(np.fabs(Ms_h/Ms-1) < 1E-6)


def test_mdef_concentration():
    # Duffy 200 matter
    cs_200m_dh = hmd_200m._get_concentration(cosmo, Ms, 1.)
    # Bhattacharya 200 matter
    cs_200m_bh = hmd_200m_b._get_concentration(cosmo, Ms, 1.)
    # Duffy 200 critical
    cs_200c_dh = hmd_200c._get_concentration(cosmo, Ms, 1.)
    # Bhattacharya 200 critical
    cs_200c_bh = hmd_200c_b._get_concentration(cosmo, Ms, 1.)
    # Klypin virial
    cs_vir_kh = hmd_vir._get_concentration(cosmo, Ms, 1.)
    # Bhattacharya virial
    cs_vir_bh = hmd_vir_b._get_concentration(cosmo, Ms, 1.)
    # Prada 200 critical
    cs_200c_ph = hmd_200c_p._get_concentration(cosmo, Ms, 1.)
    # Diemer 200 critical
    cs_200c_dih = hmd_200c_di._get_concentration(cosmo, Ms, 1.)
    assert np.all(np.fabs(cs_200m_dh/cs_200m_d-1) < 1E-6)
    assert np.all(np.fabs(cs_200c_dh/cs_200c_d-1) < 1E-6)
    assert np.all(np.fabs(cs_200m_bh/cs_200m_b-1) < 3E-3)
    assert np.all(np.fabs(cs_200c_bh/cs_200c_b-1) < 3E-3)
    assert np.all(np.fabs(cs_200c_ph/cs_200c_p-1) < 1E-3)
    assert np.all(np.fabs(cs_200c_dih/cs_200c_di-1) < 1E-2)
    assert np.all(np.fabs(cs_vir_kh/cs_vir_k-1) < 1E-3)
    assert np.all(np.fabs(cs_vir_bh/cs_vir_b-1) < 3E-3)


def test_mdef_translate_mass():
    Ms_500c_h = hmd_200m.translate_mass(cosmo, Ms, 1., hmd_500c)
    assert np.all(np.fabs(Ms_500c_h/Ms_500c-1) < 1E-6)
